import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.IOException;
import java.util.Enumeration;
import java.math.BigDecimal;

import term.*;


/** TermInterpreter interacts with the User. It provides a command line
 *  to enter terms.<BR>
 *  Other commands are:<BR>
 *  <table>
 *  <tr><th>Command</th><th>Result</th></tr>
 *  <tr>
 *    <td>!</td><td>Exit program</td>
 *  </tr>
 *  <tr>
 *    <td>[Term]:#var1:#var2: ...</td>
 *    <td>Assign the Result of the Term to variables #var1, #var2, ...<BR>
 *          Variable names generally start with the character '#'.
 *			Assignments can only occur at the end of an input line.
 *    </td>
 *  </tr>
 *  <tr>
 *    <td>#LAST</td>
 *    <td>Special variable that holds the last term that was calculated.
 *			Assignments to this variable will be accepted but are useless.
 *    </td>
 *  </tr>
 *  <tr>
 *    <td>VARS</td>
 *    <td>Returns the list of defined variables.</td>
 *  </tr>
 *  <tr>
 *    <td>SCALE(x)</td>
 *    <td>Sets the number of decimals used in calculations to x.</td>
 *  </tr>
 *  </table>
 */
public class TermInterpreter implements Runnable
{
  // Prompt:
  protected String Prompt=">";

  // momentan eingestellte Genauigkeit
  protected int actScale = 10;


	// bekannte Variablen
  protected VariableTable Vars = new VariableTable();

  // zum Lesen der Eingabezeile von StdIn
  protected BufferedReader dataIn = new BufferedReader(
								 new InputStreamReader(System.in));

/** Einlesen des Terms */
protected String inputTerm()
{
	// Promptanfrage ausgeben
	print(Prompt);
	try
	{
		// Zeile einlesen
		return dataIn.readLine().toUpperCase();
	}
	catch (IOException e)
	{
		println("Error reading the input line !");
		return "";
	}
} // inputTerm()      
/** Eintragen eines neuen Wertes in die Variablenliste */
protected void newVar(String name, String wert) throws TermFaultException
{
	newVar(name, TermBaum.parseTerm(wert,0));
} // newVar()        
/** Eintragen eines neuen Wertes in die Variablenliste */
protected void newVar(String name, TermBaum wert) throws TermFaultException
{
	if (name.length()<2)
		throw new TermFaultException("Fehler im Variablennamen");

	if (name.charAt(0) != '#')
		throw new TermFaultException("Variablennamen beginnen immer mit '#'");
	
	if (!Character.isLetter(name.charAt(1)))
		throw new TermFaultException("Fehler im Variablennamen");

	for (int i=2; i<name.length(); i++)
		if (!Character.isLetterOrDigit(name.charAt(i)))
			throw new TermFaultException("Fehler im Variablennamen");

	Vars.addVariable(name.toUpperCase(), wert);
} // newVar()        
  // Ausgaberoutinen
  protected void print(String S){ System.out.print(S); }      
  protected void println(String S){ System.out.println(S); }      
/** Bereitstellen der Oberflaeche, Term einlesen und bearbeiten lassen. */
public void run()
{
	boolean weiter=true;

	println("TermParser");
	println("\ttype \'?\' for help");

	do
	{
		String Zeile = TermBaum.removeWhitespaces( inputTerm() );

		if (Zeile.length() > 0)
		{
			if (Zeile.equals("!"))
				weiter=false;
			else if (Zeile.equals("?"))
				showHelp();
			else if (Zeile.equals("VARS"))
			{
				Enumeration E = Vars.elements();
				if (!E.hasMoreElements())
					println("\tno variables ...");
				else
				{
					while (E.hasMoreElements())
					{
						String var = (String) E.nextElement();
						println( var + " = " + Vars.getValue(var).toString() );
					}
				}
			}
			else if (Zeile.startsWith("SCALE("))
			{
				try
				{
					actScale =
						Integer.parseInt( Zeile.substring(6,Zeile.length()-1) );
				}
				catch (NumberFormatException e)
				{
					println("Error parsing the SCALE() function");
				}
			}
			else
			{
				// Term eingegeben => parsen und berechnen
				verarbeiteZeile(Zeile);
			} // else
		} // if
	} // do
	while (weiter);
} // run()
/** Display a list of valid commands */
private void showHelp()
{
	println("Valid commands:");
	println("\t!         -  exit the program");
	println("\tVARS      -  show variables");
	println("\tSCALE(x)  -  set number of digits used for calculations");
	println("");
	println("other input will be interpreted as Term");
	println("\tinclude #LAST in terms to repeat the last term that was parsed");
}
/**
 * Builds a new TermInterpreter and generates the neadfull constants
 * <code>#e</code> (Euler's constant) and <code>#pi</code> (circle constant).
 */
public TermInterpreter()
{
	newVar("#pi", String.valueOf( Math.PI ));
	newVar("#e",  String.valueOf( Math.E ));
}
// Verarbeiten einer eingegebenen Zeile, Belegung von Variablen, etc.
protected void verarbeiteZeile(String Zeile)
{
	// Variablenbelegungen aus dem Term entfernen
	String VarBel="";

	int p = Zeile.indexOf(":");

	if (p>=0)
	{
		VarBel = Zeile.substring(p);
		Zeile = Zeile.substring(0,p);
	}

	if (Zeile.length()>0) // wirklich was da zum Parsen ?
	{
		try
		{
			// Term parsen und berechnen
			TermBaum parseResult = TermBaum.parseTerm(Zeile).vereinfache();
			String result = parseResult.berechne(Vars, actScale).toString();

			// Ergebnis ausgeben
			println(parseResult.toString() + " = " + result);

			// alten Term speichern
			newVar("#LAST", parseResult);

			// Variablenzuweisung erwuenscht ?
			if (VarBel.length() > 0)
			{
				// Variablennamen bestimmen und Ergebnis als Wert zuweisen
				try
				{
					int pos=1, newPos;
					while ( (newPos=VarBel.indexOf(":",pos)) >= 0)
					{
						newVar(VarBel.substring(pos,newPos),result);
						pos=newPos+1;
					}

					newVar(VarBel.substring(pos),result);
				} // try
				catch (TermFaultException e)
				{
					println("Error assigning variables !");
				}
			}

		} // try
		catch (TermFaultException fehler)
		{
			// Fehler beim Parsen oder Berechnen aufgetreten
			if (fehler.where()>=0)
			{
				StringBuffer S = new StringBuffer();
				for (int i=fehler.where(); i-->0; S.append('-'));
				println(Zeile);
				println(S.toString()+"^ "+fehler.toString());
			} // if
			else
				println(fehler.toString());

		} // catch
		catch (StackOverflowError e)
		{
			println("Stack overflow ... recursive calculations ?");
		}
		println("");
	} // if
} // verarbeiteZeile()
} // class Interpreter
