package term;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.Enumeration;

/**
 * TermBaum of a Funktion.
 *
 * <P>This is the superclass of the implementations of various functions.
 * New functions can be added by adding new sub classes to this package.
 * </P>
 * <P>New function classes are instantiated dynamically by name if they
 * occur in the parsed function.
 * </P>
 * <table border=2>
 * <tr><th>Implemented functions</th><th>description</th></tr>
 * <tr><td>ABS(p1)</td>
 *		<td>calculates the absolute value (?)<BR>
 *			ABS(p1) = { (p1) if p1&gt;=0, (-p1) if p1&lt;0 }
 *		</td></tr>
 * <tr><td>SGN(p1)</td>
 *		<td>sign finction<BR>
 *			SGN( p1 ) = { (-1) if p1&lt;0, (0) if p1=0, (1) if p1&gt;0 }
 *		</td></tr>
 * <tr><td>CEIL(p1), FLOOR(p1)</td>
 *		<td>round up/down<BR>
 *			CEIL ( 0.1 ) = 1, FLOOR( 0.9 ) = 0
 *		</td></tr>
 * <tr><td>COS(p1), SIN(p1)</td>
 *		<td>cosine, sine<BR>
 *			Warning: the calculation of these functions is very inaccurate
 *			due to a neccessary conversion into "double" precision.
 *		</td></tr>
 * <tr><td>DIFF(v,p1)</td>
 *		<td>generates the symbolic derivation of the term p1 with respect to the variable v
 *		</td></tr>
 * <tr><td>DIV(p1,p2), MOD(p1,p2)</td>
 *		<td>integer division and modulo (?)<BR>
 *			DIV( a*b+c, b ) = a<BR>
 *			MOD( a*b+c, b ) = c<BR>
 *			Note: p1 and p2 have to be integers to apply this function !
 *		</td></tr>
 * <tr><td>EXP(p1), LN(p1)</td>
 *		<td>natural power function and logarithm
 *		</td></tr>
 * <tr><td>FAK(p1)</td>
 *		<td>factorial<BR>FAK( p1 ) = p1*(p1-1)*(p1-2)* ... * 1
 *		</td></tr>
 * </table>
 */
public abstract class Funktion extends TermBaum
{
/** Abstract method that generates the derivation of a specific
 *  function.<BR>
 *  Non differentiable functions are encouraged to throw a
 *  TermFaultException at this point.
 */
public abstract TermBaum ableitung(String variable);
/** The number of the parameters this function expects.
 *	If the number in the parsed term differes from the
 *  return value here, an Exception will be generated.
 */
protected abstract int AnzOperanden();
/** Calculate the function value for the given value of the
 *  variables. The second parameter gives the number of decimals. */
public abstract BigDecimal berechne(VariableTable variables, int Genauigkeit);
/** copies a complete TermBaum */
public TermBaum cloneTerm()
{
	try
	{
		TermBaum F = (TermBaum) this.getClass().newInstance();
		F.Operanden = (Liste)Operanden.clone();
		return F;
	}
	catch (Exception e)
	{
		return null; // "Never" occures ! ;)
	}
} // cloneTerm()
/**
 * Throws an exception to report a non differentiable function.
 */
protected final TermBaum errorNoDerivation()
{
	throw new TermFaultException( this.getClass().getName() +
			" does not have a derivation" );
}
	/** Special Constructor for subclasses with one operand. */
	protected Funktion(TermBaum operand)
	{
		super.operand = operand;
	}
/** Parses the single parameters of a function into a list. */
protected void parse2Liste(String Term, int Offset)
	throws TermFaultException
{
	int AnzPar = 0;
	int parOffset = -1;

	// split parameter at the comma characters
	BracketST Parameter = new BracketST( Term, ",", "(", ")");
	Operanden = new Liste();

	// parse each parameter
	while (Parameter.hasMoreTokens())
	{
		// calculate offset
		parOffset = Parameter.getNextPos();

		String next = Parameter.nextToken();
		debugP("next parameter: "+next);

		// parse one parameter and add it's TermBaum to the parameter list
		Operanden.addLast( parseTerm(next, Offset+parOffset) );
		AnzPar++;

		// don't care about the commas themselves
		if (Parameter.hasMoreTokens())
			Parameter.nextToken();
	} // while

	if (AnzPar != AnzOperanden())
	{
		throw new TermFaultException(Offset,
			 "Function expects exactly "+AnzOperanden()+" Parameter(s) !");
	}
	else if (AnzPar == 1)
	{
		operand = (TermBaum) Operanden.firstEl();
		Operanden = null;
	}
} // parse2Liste()
/** Parse a Term like 'F(p1,p2,...)',
 *	returning an object representing the function that was
 *  recognized.<BR>
 *  New functions can be added by writing sub classes of this one
 *  that will be instantiated by name if they occur in the parsed
 *  function.
 */
public static Funktion parseFunktion(String Term, int Offset)
	throws TermFaultException
{
	int nameEnd = Term.indexOf('(');
	if (nameEnd < 0)
		throw new TermFaultException(Offset,"Function without parameters");

	if (!Term.endsWith(")"))
		throw new TermFaultException(Offset+nameEnd,"malformed parameterlist");

	// Namens- und Parameterteil der Funktion teilen
	String Name = Term.substring(0,nameEnd);
	String ParameterString = Term.substring(nameEnd+1,Term.length()-1);

	debugP("generating Function: "+Name);

	Funktion newFkt;
	try
	{
		newFkt = (Funktion) Class.forName(packagePraefix+Name).newInstance();
	}
	catch (Exception e)
	{
		throw new TermFaultException(Offset,"unknown Function: "+Name);
	}

	newFkt.parse2Liste(ParameterString, Offset+nameEnd+1);

	return newFkt;
} // parseFunktion()
/** Returns a String representatin of the function and its parameters. */
public String toString()
{
	StringBuffer term;
	String fktName = getClass().getName();

	int pos = fktName.lastIndexOf(".");
	if (pos >= 0)
		term = new StringBuffer( fktName.substring(pos+1) );
	else
		term = new StringBuffer(fktName);

	term.append('(');

	if (AnzOperanden() > 1)
	{
		Enumeration param = Operanden.elements();
		if (param.hasMoreElements())
		{
			term.append( ((TermBaum) param.nextElement()).toString() );

			while (param.hasMoreElements())
				term.append(',').append( ((TermBaum) param.nextElement()).toString() );
		}
	}
	else
	{
		// single parameters are accessed by getOperand()
		term.append( getOperand().toString() );
	}

	term.append(')');

	return term.toString();
} // toString()
	/** protected Constructor for subclasses */
	protected Funktion(){}
/** Sub classes should return a new TermBaum
 *  that is simplified as much as possible.
 */
public abstract TermBaum vereinfache();
} // class Funktion
