package term;

import java.math.BigDecimal;
import java.math.BigInteger;

import java.util.Enumeration;

/**
 * TermBaum is the super class of all parts of the Tree representation
 * of the Term. It integrates basic parse methods to generate a TermBaum
 * structure from a given function string. Subclasses have to provide
 * specific methods (abstract here) to implement unique functions.<BR>
 * @see #ableitung(java.lang.String)
 * @see #berechne(term.VariableTable,int)
 * @see #cloneTerm()
 * @see #toString()
 * @see #vereinfache()
 */
public abstract class TermBaum
{
	/**  Name of the package that keeps the function classes */
	protected final static String packagePraefix = "term.";

	/** ParseDebug = true outputs debug messages during parsing */
	protected static final boolean ParseDebug = true;

	/** CalcDebug = true outputs debug messages during calculations */
	protected static final boolean CalcDebug = true;

	/** StrictSyntax = true forbids free signs in sums, ... */
	protected static final boolean StrictSyntax = false;


	/** The parameters of this element of the TermBaum stucture */
	protected Liste Operanden = null;

	/** The parameter if it is only a single operand. */
	protected TermBaum operand = null;

	/** Calculation symbols ordered by priority */
	protected final static char[] rz  = {'+', '-', '*', '/', '^'};
	/** Build the Deviation with respect to the given variable.
	 */
	public abstract TermBaum ableitung(String variable);
	/** Calculates the (recursively calculated) value of this TermBaum.
	 */
	public abstract BigDecimal berechne(VariableTable Variablen, int Genauigkeit);
	/** If Variables are not needed this method is a bit shorter to be called. */
	public static BigDecimal berechneTerm(String Zeile, int Scale)
		throws TermFaultException
	{
		return berechneTerm(Zeile, new VariableTable(), Scale);
	} // berechneTerm()
	/** Calculating the value of a given string term considering the given variables. */
	public static BigDecimal berechneTerm(String Zeile, VariableTable Vars, int Scale)
		throws TermFaultException
	{
		// parse and calculate the Term
		TermBaum T = parseTerm(Zeile,0).vereinfache();

		debugP("simplified Term: " + T.toString());

		BigDecimal Wert = T.berechne(Vars, Scale);

		return Wert;
	} // berechneTerm()
	/** returns a copy of this TermBaum as an Object. */
	public Object clone()
	{
		return cloneTerm();
	}
	/** Builds a complete recursive copy of this TermBaum.
	 */
	public abstract TermBaum cloneTerm();
	// debug messages during calculations:
	protected static void debugC(String M)
	{
		if (CalcDebug) System.out.println(M);
	}
	// debug messages during parsing:
	protected static void debugP(String M)
	{
		if (ParseDebug) System.out.println(M);
	}
	/**
	 * Insert the method's description here.
	 * Creation date: (28.11.1999 %T)
	 * @return java.lang.String[]
	 */
	public String[] getDependencies()
	{
		Liste vars = new Liste();
		getDependenciesRec(vars);

		String[] varNames = new String[ vars.size() ];

		Enumeration names = vars.elements();

		int i=0;	
		while (names.hasMoreElements())
			varNames[i++] = (String) names.nextElement();
			
		return varNames;
	}
	/**
	 * Insert the method's description here.
	 * Creation date: (28.11.1999 %T)
	 * @return java.lang.String[]
	 */
	protected void getDependenciesRec(Liste list)
	{
		if (this instanceof Variable)
		{
		}
		else if (operand != null)
		{
		}
		else
		{
		}
	}
	/** Looks for the weakest calculation symbol in the top most
	 *  level of the Term.
	 *  @return the index of the symbol in the "rz" array or -1
	 *     if no symbol was found.
	 */
	public static int getMaxRZ(String Term)
	{
		// initially chose minimum beyond array limit
		int MinIndex=rz.length;

		// skip the sign of the first operand
		if (Term.charAt(0)=='-' || Term.charAt(0)=='+')
			Term = Term.substring(1);

		// only search the first level, cut at symbol characters
		BracketST ST = new BracketST(Term, new String(rz), "(", ")");

		// search the weakest symbol
		while (ST.hasMoreTokens() && MinIndex>0)
		{
			String next = ST.nextToken();

			if (next.length()==1)
			{
				// only one character => maybe a calculation symbol ?
				char c = next.charAt(0);
				for (int i=MinIndex-1; i>=0; i--)
				{
					if (c==rz[i])
						MinIndex = i;
				}
			} // if
		} // while

		if (MinIndex<rz.length)
			return MinIndex;
		else
			return -1;
	} // getMaxRZ()
	/** Returns the value operand field, i.e. the operand if only one exists.
	 *  Returns null if the Symbol / Function / ... has more or less than one
	 *  operand.
	 */
	protected TermBaum getOperand()
	{
		return operand;
	}
/**
 * Insert the method's description here.
 * Creation date: (03.12.1999 %T)
 * @param operanden term.Liste
 */
protected TermBaum(Liste operanden)
{
	this.Operanden = operanden;
}
/**
 * Insert the method's description here.
 * Creation date: (03.12.1999 %T)
 * @param operand term.TermBaum
 */
protected TermBaum(TermBaum operand)
{
	this.operand = operand;
}
	/** Remove outer brackets. The Term must not have any Whitespaces.
	 *  @see #removeWhitespaces(java.lang.String)
	 */
	public static String ohneKlammernAussen(String Term, int Offset)
		throws TermFaultException
	{
		while ( Term.startsWith("(") &&
				sucheKlammerEnde(Term,0,Offset) == Term.length()-1 )
		{
			Term = Term.substring(1,Term.length()-1);
		} // while
		return Term;
	} // ohneKlammernAussen()
	/** Parses a Term into a Tree representation.<BR>
	 *  @param Offset the current character offset within the input line.
	 *  @param Term the String that will be parsed.
	 *  The String must not include any Whitespaces. This can be
	 *  achieved using the method removeWhitespaces().
	 *  @see #removeWhitespaces(java.lang.String)
	 */
	public static TermBaum parseTerm(String Term, int Offset)
		throws TermFaultException
	{
		// remove outer brackets
		String newTerm = ohneKlammernAussen(Term, Offset);

		Offset += Term.indexOf(newTerm); // rearrange the global offset
		Term = newTerm;

		if (Term.length()>0)
		{
			debugP("Parsing Term: " + Term);

			// find weakest calculation symbol in the Term
			int RzNr = getMaxRZ(Term);

			if (RzNr>=0)
			{
				// the Term can be divided at the symbol characters
				return teileNachRZ(Term,Offset,rz[RzNr]);
			}
			else
			{
				// no calculation symbols found
				switch (Term.charAt(0))
				{
					case '-':
						// negative signs are represented separately
						return new Negativ(Term.substring(1), Offset+1);
					case '+':
						// positive signs can be left out
						return parseTerm(Term.substring(1),Offset+1);
					case '#':
						// Variable found
						return new Variable(Term,Offset);
					default:
						if (Character.isDigit(Term.charAt(0)))
						{
							// first character is a digit => number found !
							return new Konstante(Term,Offset);
						}
						else
						{
							// no digit found => try to parse a function
							return Funktion.parseFunktion(Term,Offset);
						}
				} // switch
			} // else
		} // if
		else
			throw new TermFaultException(Offset,"missing Term");
	} // parseTerm()
	/** Parses a complete Term line into a TermBaum structure.<BR>
	 *  @param Offset the current character offset within the input line.
	 *  @param Term the String that will be parsed.
	 *  The String must not include any Whitespaces. This can be
	 *  achieved using the method removeWhitespaces().
	 *  @see #removeWhitespaces(java.lang.String)
	 */
	public final static TermBaum parseTerm(String term)
	{
		return parseTerm(term, 0);
	}
	/** Returns a copy of the String that has all Whitespaces removed.
	 */
	public static String removeWhitespaces(String term)
	{
		StringBuffer newTerm = new StringBuffer( term.length() );

		// append all characters but whitespaces
		for (int i=0; i<term.length(); i++)
		{
			char nextChar = term.charAt(i);
			if ( ! Character.isWhitespace(nextChar) )
				newTerm.append(nextChar);
		}

		return newTerm.toString();
	}
	/** Look for the bracket terminating the bracket Term beginning
	 *  at the given location.
	 */
	public static int sucheKlammerEnde(String Term, int Pos, int Offset)
		throws TermFaultException
	{
		// count bracket levels skipping the first opening bracket
		int Klammer=1;

		for (int p=Pos+1; p<Term.length(); p++)
		{
			switch (Term.charAt(p))
			{
				case '(':
					Klammer++; break;
				case ')':
					Klammer--;
					if (Klammer==0)
						return p;
					break;
			} // switch
		} // for

		// if the code arrives here, the bracket syntax is malformed: Klammer>0 !
		throw new TermFaultException(Pos+Offset,"Missing terminating bracket !");
	} // sucheKlammerEnde()
	/** Divides the Term by the given calculation symbol.
	 *  @return the parsed TermBaum (sum or product).
	 */
	protected static TermBaum teileNachRZ(String Term, int Offset, char symbol)
		throws TermFaultException
	{
		switch (symbol)
		{
			case '+':
			case '-':
				return new Summe(Term, Offset);
			case '*':
			case '/':
				return new Produkt(Term, Offset);
			case '^':
				return new Power(Term, Offset);
			default:
				return null; // shouldn't occur !
		} // switch
	} // teileNachRZ()
	/** Returns a String representation of this TermBaum.
	 *  Therefore the sub classes should return the represented Term
	 *  in infix notation.
	 */
	public abstract String toString();
/**
 * Insert the method's description here.
 * Creation date: (03.12.1999 %T)
 */
protected TermBaum() {}
	/** Simplifying the Term.
	 */
	public abstract TermBaum vereinfache();
} // class TermBaum
